/*
** hash.c - hash table stuff
**
** Keith Zaback (krz@wcug.wwu.edu) Apr 98
*/

#include "ipxfury.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* options */
#define SHOW_INSERTS	0


/* protos */
void		empty_bucket(HASHENTRY *entry);

/*
**	HASH_START_TRAV - starts a traversal
*/
void *hash_start_trav(HASHTABLE *htab, HASHTRAV *trav)
{
	int		i;
	
	/* ok.. build the struct */
	trav->htab = htab;
	
	/* find 1st item */
	for( i = 0; i < htab->total_buckets; i++ ) {
		/* do we have something? */
		if( htab->table[i] != NULL ) {
			/* fill out the rest */
			trav->bucket	= i;
			trav->entry	= htab->table[i];
			return(trav->entry->data);
		}
			
	}
	
	/* nothing */
	return(NULL);

}


/*
**	HASH_NEXT_TRAV - returns the next item
*/
void *hash_next_trav(HASHTRAV *trav)
{
	int		i;
	/* any more in this bucket? */
	if( trav->entry->next != NULL ) {
		/* move to the next one */
		trav->entry = trav->entry->next;
		return(trav->entry->data);
	}
	
	/* that bucket was empty.. find the next */
	for( i = (trav->bucket + 1); i < trav->htab->total_buckets; i++ ) {
		if( trav->htab->table[i] != NULL ) {
			/* give them this */
			trav->bucket	= i;
			trav->entry	= trav->htab->table[i];
			return(trav->entry->data);
		}
	}
	
	/* nothing more */
	return(NULL);
}



/*
**	HASH_REMOVE_MULTI - remove multi things in one pass..
*/
int hash_remove_multi(HASHTABLE *htab, int (*testfunc)(void *))
{
	HASHENTRY	*entry, *shadow, *remove;
	int		i;
	
	/* make sure this is a hashtable */
	if( htab->magic != HASH_MAGIC ) {
		errormsg(0, "hash_remove_multi", "bad hash magic");
		return(0);
	}
	
	/* for each entry call the enum function */
	for( i = 0; i < htab->total_buckets; i++ ) {
		/* does this bucket contain anything? */
		if( htab->table[i] != NULL ) {
			/* yes, so step thru the entries */
			shadow	= NULL;
			entry	= htab->table[i];
			while( entry != NULL ) {
				if( testfunc(entry->data) ) {
					/* delete this one */
					if( shadow == NULL ) {
						htab->table[i] = entry->next;
						remove = entry;
						entry = entry->next;
					} else {
						shadow->next = entry->next;
						remove = entry;
						entry = entry->next;
					}
					
					/* free mem */
					if( remove->data != NULL )
						free(remove->data);
					free(remove);
					
					/* dec entries */
					htab->entries--;
					
				} else {
					/* move to the next one */
					shadow	= entry;
					entry	= entry->next;
				}
			}
		}
	}
	
	/* all done */
	return(1);


}


/*
**	HASH_EMPTY_TABLE - empties a hash table
*/
void hash_empty_table(HASHTABLE *htab)
{
	int		i;
	
	/* make sure it is ok */
	if( htab == NULL )
		return;
		
	/* free the alloc'd mem */
	if( htab->table != NULL ) {
		for( i = 0; i < htab->total_buckets; i++ ) 
			empty_bucket(htab->table[i]);
	}
	
	/* just clean out the memory */
	memset(htab->table, 0, (sizeof(void *) * htab->total_buckets));
	
	/* reset the counters */
	htab->entries = 0;
	
	/* done */

}



/*
**	HASH_ENUM_ENTRIES()
*/
void hash_enum_entries(HASHTABLE *htab, void (*enumfunc)(int, void *))
{
	HASHENTRY	*entry;
	int		i;
	
	/* make sure this is a hashtable */
	if( htab->magic != HASH_MAGIC ) {
		errormsg(0, "hash_enum_entries", "bad hash magic");
		return;
	}
	
	/* for each entry call the enum function */
	for( i = 0; i < htab->total_buckets; i++ ) {
		/* does this bucket contain anything? */
		if( htab->table[i] != NULL ) {
			/* yes, so step thru the entries */
			entry = htab->table[i];
			while( entry != NULL ) {
				enumfunc(i, entry->data);	/* bucket # and data */
				entry = entry->next;		/* move to next entry */
			}
		}
	}
	
	/* all done */
	return;
}



/*
**	HASH_GET()
*/
void *hash_get(HASHTABLE *htab, void *item)
{
	int		bno;
	HASHENTRY	*entry;
	
	/* check it */
	if( htab->magic != HASH_MAGIC ) {
		errormsg(0, "hash_get", "bad hash magic!");
		return(NULL);
	}
	
	/* calc the bucket number */
	bno = htab->hash(item) % htab->total_buckets;

	/* go there */
	entry = htab->table[bno];
	
	/* lets try to find it */
	while( entry != NULL ) {
		if( htab->compare(entry->data, item) == 0 )
			return(entry->data);	/* found it */
		entry = entry->next;		/* next entry in this bucket */
	}
	
	/* failed */
	return(NULL);

}

/*
**	HASH_REMOVE() - removes an entry
*/
int hash_remove(HASHTABLE *htab, void *item)
{
	int		bno;
	HASHENTRY	*shadow, *ptr;
	
	/* check it */
	if( htab->magic != HASH_MAGIC ) {
		errormsg(0, "hash_remove", "bad hash magic!");
		return(0);
	}
	
	/* get bucket number */
	bno = htab->hash(item) % htab->total_buckets;
	
	/* set the ptrs */
	shadow	= NULL;
	ptr	= htab->table[bno];
	
	/* here we go */
	while( ptr != NULL ) {
		/* did we find what we needed? */
		if( htab->compare(item, ptr->data) == 0 ) {
			/* found it -- so remove it */
			if( shadow == NULL ) {
				/* 1st item */
				htab->table[bno] = ptr->next;
			} else {
				/* nth item */
				shadow->next = ptr->next;
			}
			
			/* remove */
			if( ptr->data != NULL )
				free(ptr->data);
			free(ptr);
			
			/* dec entries */
			htab->entries--;
			
			/* did ok */
			return(1);
			
		} else {
			/* nope.. */
			shadow	= ptr;
			ptr 	= ptr->next;
		}
	}
	
	/* failed */
	return(0);

}


/*
**	HASH_INSERT()
*/
void *hash_insert(HASHTABLE *htab, void *item)
{
	int		bno;
	HASHENTRY	*temp;
	
	/* check it */
	if( htab->magic != HASH_MAGIC ) {
		errormsg(0, "hash_insert", "bad hash magic!");
		return(NULL);
	}

#if SHOW_INSERTS == 1
	notice("hash_insert:\n");
	hexdump(item, htab->entry_size);
#endif

	/* allocate and create a new entry */
	temp = (HASHENTRY *)malloc(sizeof(HASHENTRY));
	if( temp == NULL ) {
		errormsg(0, "hash_insert", "failed to allocate new entry");
		return(NULL);	/* failed */
	}
	
	/* alloc a place for the data */
	temp->data = (void *)malloc(htab->entry_size);
	if( temp->data == NULL ) {
		errormsg(0, "hash_insert", "could not alloc space for item (%d bytes)", htab->entry_size);
		free(temp);
		return(NULL);
	}
	
	/* copy the data item */
	memcpy(temp->data, item, htab->entry_size);
	
	/* figure out what bucket */
	bno = htab->hash(item) % htab->total_buckets;
		
	/* insert it into the bucket */
	temp->next = htab->table[bno];
	htab->table[bno] = temp;
	
	/* one more entry in the table */
	htab->entries++;
	
	/* return a pointer to the new entry's data */
	return(temp->data);

}



/*
**	CREATE_HASH_TABLE()
*/
int create_hash_table(	HASHTABLE 	*htab,
			int 		total_entries,
			int 		entry_size,
			int 		(*hash)(void *),
			int 		(*compare)(void *, void *))
{
	int		alloc_size;
	
	/* calc */
	alloc_size = total_entries * sizeof(void *);

	/* set the total entries & size of entry */
	htab->total_buckets = total_entries;	/* set max size */
	htab->entries = 0;			/* no entries yet */
	htab->compare = compare;
	htab->hash = hash;
	htab->entry_size = entry_size;
	
	/* alloc memory for the table data */
	htab->table = (HASHENTRY **)malloc(alloc_size);
	if( htab->table != NULL ) {
		memset(htab->table, 0, alloc_size);	/* clean it */
		htab->magic = HASH_MAGIC;		/* set this */
		return(1);
	}

	/* failed */	
	errormsg(0, "create_hash_table", "could not allocate table memory (%d entries)", total_entries);
	return(0);
	
}


/*
**	FREE_HASH_TABLE()
*/
void free_hash_table(HASHTABLE *htab)
{
	int		i;

	/* make sure it is valid */
	if( htab == NULL ) {
		errormsg(0, "free_hash_table", "hash table is NULL");
		return;
	}
	
	/* check it */
	if( htab->magic != HASH_MAGIC ) {
		errormsg(0, "free_hash_table", "bad hash magic!");
		return;
	}
	
	/* free the alloc'd mem */
	if( htab->table != NULL ) {
		for( i = 0; i < htab->total_buckets; i++ ) 
			empty_bucket(htab->table[i]);
		free(htab->table);
	}
		
	/* done */
	return;
}


/*
**	EMPTY_BUCKET()
*/
void empty_bucket(HASHENTRY *entry)
{
	HASHENTRY	*ptr, *shadow;
	
	/* init */
	shadow = NULL;
	ptr = entry;		
	while( ptr != NULL ) {
		if( ptr->data != NULL )		/* free their data */
			free(ptr->data);
		shadow = ptr;
		ptr = ptr->next;		/* move along */
		free(shadow);			/* free this */
	}
	
}

