/*
** serial.c - this is for serial type actions
**
** Keith Zaback (krz@wcug.wwu.edu) Mar 98
*/

#include "ipxfury.h"
#include <stdio.h>
#include <termios.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <netinet/in.h>

/* defines */
#define SERIAL_WARNINGS		0

/* extern */
extern GLOBALS	ifg;

/* protos */
int		termios_baud(int baud);

/* globals */
BYTE		serial_buffer[FURY_SERIAL_TRANS_BUF_SIZE];


/*
**	SERIAL_QUEUE_IPX_FRAME - this is called by the "opposite"
**				 device to send an ipx frame out.
**				 This frame is in network byte order
*/
int serial_queue_ipx_frame(INTERFACE *from, INTERFACE *ser, BYTE *ipxframe)
{
	BYTE		new_fury_frame[FURY_FRAME_MAX];
	int		new_fury_length;
	int		ftype;
	
	/* no loops! */
	if( from == ser ) {
		notice("[%s] no loops\n", ser->name);
		return(0);
	}

	/* build a fury'd ipx frame */
	if( ! create_fury_ipx_frame(ser, ipxframe, new_fury_frame, &ftype, &new_fury_length) ) {
		return(0);
	}
	
	/* ready to send it out! */
	send_fury_frame(ser, ftype, new_fury_frame, new_fury_length);
	
	/* all done */
	return(1);
	
}

/*
**	SERIAL_PREPARE_DEVICE
*/
int serial_prepare_device(INTERFACE *ser)
{
	int		modem_lines;
	
	/* look for serial device, and if carrier send it the UP event */
	if( ser != NULL ) {
		if( serial_status(ser, &modem_lines) ) {
			if( (modem_lines & FURY_MODEM_CARRIER) == FURY_MODEM_CARRIER ) {

				/* send the UP event to this guy */
				ser->event_handler(ser, FURY_EVENT_UP);
				ser->event_handler(ser, FURY_EVENT_OPEN);
			} else {
				errormsg(0, "ipx_fury", "carrier not detected on '%s'", ser->name);
				ser->event_handler(ser, FURY_EVENT_DOWN);
			}
		}
	}

	/* done */
	return(1);

}

/*
**	SERIAL_FLUSH - should flush the input buffer for the device..
*/
int serial_flush(INTERFACE *ser)
{
	int		i, nfds;
	fd_set		readfds;
	char		buffer[512];
	struct timeval	tv;
	int		bytes;
	
	/* init the counter */
	bytes = 0;
	i = 0;
	while( i < SERIAL_FLUSH_MAX ) {
		/* clean fdset */
		FD_ZERO(&readfds);
		
		/* set this device */
		FD_SET(ser->fd, &readfds);
		
		/* timeout */
		tv.tv_sec = 0;
		tv.tv_usec = 250000;	/* 1/4 second */
		
		/* now select */
		nfds = select(FD_SETSIZE, &readfds, NULL, NULL, &tv);
		if( nfds == 0 ) {
			/* timeout */
			break;
		} else if( nfds > 0 ) {
			/* read in the packet */
			bytes += read(ser->fd, buffer, sizeof(buffer));
		} else {
			errormsg(0, "serial_flush", "select() error");
			break;
		}
		/* inc counter */
		i++;
	}
#if 0	
	/* info */
	errormsg(0, "serial_flush", "removed %d bytes", bytes);
#endif	
	/* done */
	return(1);
	
}



/*
**	SERIAL_TIMEOUT - handles timeout events, for the LCP
**
*/
int serial_timeout(INTERFACE *ser)
{
#if 0
	errormsg(0, "serial_timeout", "timeout!");
#endif

	/* dec the counter */
	ser->restart_counter--;

	/* check the restart counter */
	if( ser->restart_counter > 0 ) {
		/* send the event for counter > 0 */
		ser->event_handler(ser, FURY_EVENT_TON);
		/* add an event */
		add_event(ifg.events, ETYPE_FLCP_TIMEOUT, 2, 0, (void (*)(void *, int))ser->timeout_handler, (void *)ser);
	} else {
		/* send the event for counter == 0 */
		ser->event_handler(ser, FURY_EVENT_TOZ);
	}
	
	/* everything looks ok, so lets return success	*/
	return(1);
}


/*
**	SERIAL_STATUS - returns line status for modems, and co
*/
int serial_status(INTERFACE *ser, int *result)
{
	
	/* make the ioctl call */
	if( ioctl(ser->fd, TIOCMGET, result) == 0 ) {
		/* did fine */
		return(1);
	}
		
	/* failed */
	return(0);
}


/*
**	SERIAL_WRITE - this writes all the data in sendbuf out via the modem
*/
int serial_write(INTERFACE *ser)
{
	BYTE		*data;
	int		length;
	int		bytes_out;

	/* init */
	length = 0;
	
	/* get pointers to data */
	buffer_get_all(&ser->sendbuf, &data, &length);
	
	/* anything to really send? */
	if( length == 0 ) {
		return(0);
	}
	
	/* ok, there IS something to send */
	bytes_out = write(ser->fd, data, length);
	if( (bytes_out != length) && (bytes_out >= 0) ) {
#if SERIAL_WARNINGS == 1
		errormsg(0, "serial_write", "problem sending data (%d/%d bytes)", bytes_out, length);
#endif	
		/* we didn't send it all, so put back what we couldn't send out */
		data	+= bytes_out;
		length	-= bytes_out;
		buffer_insert(&ser->sendbuf, data, length);
		/* set the flags */
		ser->flags |= FURY_IF_TOO_MUCH_DATA;
	} else if( bytes_out < 0 ) {
#if SERIAL_WARNINGS == 1
		errormsg(0, "serial_write", "error sending data!");
#endif
		/* put it all back */
		buffer_insert(&ser->sendbuf, data, length);
		bytes_out = 0;
	} else {
		/* remove the too much data flag */
		ser->flags &= ~FURY_IF_TOO_MUCH_DATA;
	}
	
	/* adjust stats */
	ser->stats.send_bytes_recent 	+= bytes_out;
	ser->stats.bytes_sent		+= bytes_out;
	
	/* done */
	return(1);
}





/*
**	SERIAL_PROCESS - processes the recvbuf for the serial device
*/
int serial_process(INTERFACE *ser)
{
	BYTE		*data;
	int		length;
	
	/* init */
	length = 0;
	
	/* get all the data in the buffer */
	buffer_get_all(&ser->recvbuf, &data, &length);
	
	/* is there something to process? */
	if( length == 0 ) {
		return(0);
	}

	/* set the number of frames sent to zero for this tick */
	ser->frames_sent = 0;
	
	/* now process the data */
	fury_receive(ser, data, length);
	
	return(1);
}




/*
**	SERIAL_READ - reads data from the serial device and puts it in the
**                    recvbuf
*/
int serial_read(INTERFACE *ser)
{
	int		bytes_read;
	
	/* read the data from the modem */
	bytes_read = read(ser->fd, serial_buffer, FURY_SERIAL_TRANS_BUF_SIZE);

	/* did we get anything useful? */
	if( bytes_read > 0 ) {
		/* adjust stats */
		ser->stats.recv_bytes_recent 	+= bytes_read;
		ser->stats.bytes_rcvd		+= bytes_read;
		
		/* add this data into the recvbuf */
		if( buffer_insert(&ser->recvbuf, serial_buffer, bytes_read) ) {
			/* we are done, just return # of bytes read */
			return(bytes_read);		
		} else {
			/* could not insert data */
			errormsg(0, "serial_read", "could not insert %d bytes into recvbuf", bytes_read);
			return(0);
		}
	}

	/* failed */	
	return(0);
}



/*
**	SERIAL_OPEN
*/
int serial_open(INTERFACE *ser)
{
	char		filename[64];
	struct termios	tio;
	int		baud;
	
	/* create filename by adding "/dev" */
	sprintf(filename, "/dev/%s", ser->name);
	
	/* now open this up */
	ser->fd = open(filename, (O_RDWR | O_NOCTTY | O_NONBLOCK));
	if( ser->fd < 0 ) {
		errormsg(0, "serial_open", "failed to open %s", filename);
		return(0);	
	}

	/* now get the terminal attributes, and save them for restoring */
	if( tcgetattr(ser->fd, &ser->data.s.s_tio) < 0 ) {
		errormsg(0, "serial_open", "tcgetattr() failed!");
		close(ser->fd);
		return(0);
	}

	/* copy the termios into a temp area to work with */	
	memcpy(&tio, &ser->data.s.s_tio, sizeof(tio));

	/* set the modes we want */
	tio.c_iflag &= ~(IGNBRK | INPCK | ISTRIP | INLCR | IGNCR |
	                 ICRNL | IXON | IXANY | IXOFF | IMAXBEL);
	tio.c_iflag |= (BRKINT | IGNPAR);
	tio.c_oflag &= ~(OPOST);
	tio.c_lflag &= ~(ISIG | ICANON | ECHO | IEXTEN);
	tio.c_cflag &= ~(CSTOPB | PARENB);
	tio.c_cflag |= (CS8 | CREAD | HUPCL | CRTSCTS);
	
	/* now baud rate */
	baud = termios_baud(ser->data.s.s_baudrate);		/* get proper constant */
	cfsetispeed(&tio, baud);
	cfsetospeed(&tio, baud);
	
	/* now set these attributes */
	if( tcsetattr(ser->fd, TCSANOW, &tio) < 0 ) {
		errormsg(0, "serial_open", "tcsetattr() failed!");
		close(ser->fd);
		return(0);
	}

	/* create the buffers */
	if( ! create_buffer(&ser->recvbuf, FURY_SERIAL_RECV_BUF_SIZE) ) {
		close(ser->fd);
		return(0);
	}
	
	if( ! create_buffer(&ser->sendbuf, FURY_SERIAL_TRANS_BUF_SIZE) ) {
		close(ser->fd);
		return(0);
	}
	
	/* set outgoing buffer ptr */
	ser->outbuf = &ser->sendbuf;
	
	/* init the caches */
	if( ! create_hdrcache(&ser->data.s.s_send_hdrcache, HDRCACHE_DEFAULT_SIZE) ) {
		close(ser->fd);
		return(0);
	}
	if( ! create_hdrcache(&ser->data.s.s_recv_hdrcache, HDRCACHE_DEFAULT_SIZE) ) {
		close(ser->fd);
		return(0);
	}
	
	
	/* create the pkt caches */
	if( ! create_hash_table(&ser->data.s.s_send_pktcache,
				FURY_PKTCACHE_TABLE_SIZE,
				sizeof(PKTCACHE_ENTRY),
				(int (*)(void *))pktcache_hash,
				(int (*)(void *, void *))pktcache_compare) ) {
		close(ser->fd);
		return(0);
	}
	if( ! create_hash_table(&ser->data.s.s_recv_pktcache,
				FURY_PKTCACHE_TABLE_SIZE,
				sizeof(PKTCACHE_ENTRY),
				(int (*)(void *))pktcache_hash,
				(int (*)(void *, void *))pktcache_compare) ) {
		close(ser->fd);
		return(0);
	}

	/* drop until the 1st flag */	
	memset(&ser->furyframe, 0, sizeof(FURYFRAME));
	ser->furyframe.drop = 1;
	
	/* setup cache ptrs! */
	ser->cache_ptrs.send_hdrcache	= &ser->data.s.s_send_hdrcache;
	ser->cache_ptrs.recv_hdrcache	= &ser->data.s.s_recv_hdrcache;
	ser->cache_ptrs.send_pktcache	= &ser->data.s.s_send_pktcache;
	ser->cache_ptrs.recv_pktcache	= &ser->data.s.s_recv_pktcache;
	ser->cache_ptrs.last_senddata	= ser->data.s.s_last_senddata;
	ser->cache_ptrs.last_sendsize	= &ser->data.s.s_last_sendsize;
	ser->cache_ptrs.last_recvdata	= ser->data.s.s_last_recvdata;
	ser->cache_ptrs.last_recvsize	= &ser->data.s.s_last_recvsize;
	ser->cache_ptrs.diff_reset	= &ser->data.s.s_diff_reset;
		               		
	/* everything went ok */
	ser->flags |= FURY_IF_VALID;		/* this interface is valid */
	
	/* send event to fury */
	fury_event(ser, FURY_EVENT_OPEN);	/* we have opened the device */
	
	/* set the stats recent time */
	gettimeofday(&ser->stats.recent_stats_start, NULL);
	
	/* done */
	return(1);
	
}


/*
**	SERIAL_CLOSE
*/
int serial_close(INTERFACE *ser)
{
	int		modem_lines;
	struct termios	tio, tio_save;
	
	/* check the device */
	if( (ser->flags & FURY_IF_VALID) != FURY_IF_VALID ) {
		errormsg(0, "serial_close", "called with an invalid interface.");
		return(0);
	}

	/* free the buffers */
	free_buffer(&ser->recvbuf);
	free_buffer(&ser->sendbuf);
	
	/* free the caches */
	free_hdrcache(&ser->data.s.s_send_hdrcache);
	free_hdrcache(&ser->data.s.s_recv_hdrcache);
	
	/* free the hash tables (pktcache) */
	free_hash_table(&ser->data.s.s_send_pktcache);
	free_hash_table(&ser->data.s.s_recv_pktcache);
	
	/* check the modem lines */
	if( ser->flags & FURY_IF_MODEM_LINES ) {
		if( serial_status(ser, &modem_lines) ) {
			if( (modem_lines & FURY_MODEM_CARRIER) == FURY_MODEM_CARRIER ) {
				/* get the attributes */
				tcgetattr(ser->fd, &tio);
				
				/* copy it */
				memcpy(&tio_save, &tio, sizeof(tio));
				
				/* set baud rate to B0 = hangup */
				cfsetispeed(&tio, B0);
				cfsetospeed(&tio, B0);
		
				/* tell them we are hanging up.. */
				notice("[%s] Hanging up...\n", ser->name);
				
				/* set these new attributes */
				tcsetattr(ser->fd, TCSANOW, &tio);
			
				/* wait for a couple seconds */
				sleep(3);
				
				/* restore attributes */
				tcsetattr(ser->fd, TCSANOW, &tio_save);
				
			}
		}
	}
			
	/* restore the old attributes */
	/* I used to check the return value, but it would ALWAYS	*/
	/* fail, so I am not sure what to do here.  However the modem	*/
	/* does get reset and everything appears to be fine..		*/
	
	tcsetattr(ser->fd, TCSANOW, &ser->data.s.s_tio);

	/* close this descriptor */
	alarm(2);
	close(ser->fd);
	alarm(0);
	
	/* did ok */
	return(1);
}


/*
**	TERMIOS_BAUD() - returns termios defines for the baud rate
*/
int termios_baud(int baud)
{
	int		res;

	switch( baud ) {
		case 9600:	res = B9600;	break;
		case 19200:	res = B19200;	break;
		case 38400:	res = B38400;	break;
		case 57600:	res = B57600;	break;
		case 115200:	res = B115200;	break;
		default:	res = B38400;	break;
	}
	
	return(res);
}

/*
**	SERIAL_DUMMY
*/
int serial_dummy(INTERFACE *fdev)
{
	return(1);
}